# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Creation of mimeograms. '''
# TODO? Use BSD sysexits.


from __future__ import annotations

from . import __
from . import interfaces as _interfaces
from . import tokenizers as _tokenizers


_scribe = __.produce_scribe( __name__ )


class Command(
    _interfaces.CliCommand,
    decorators = ( __.standard_dataclass, __.standard_tyro_class ),
):
    ''' Creates mimeogram from filesystem locations or URLs. '''

    sources: __.typx.Annotated[
        __.tyro.conf.Positional[ list[ str ] ],
        __.typx.Doc( ''' Filesystem locations or URLs. ''' ),
        __.tyro.conf.arg( prefix_name = False ),
    ]
    clip: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.typx.Doc( ''' Copy mimeogram to clipboard. ''' ),
        __.tyro.conf.arg( aliases = ( '--clipboard', '--to-clipboard' ) ),
    ] = None
    count_tokens: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.typx.Doc( ''' Count total tokens in mimeogram. ''' ),
    ] = None
    edit: __.typx.Annotated[
        bool,
        __.typx.Doc( ''' Spawn editor to capture introductory message. ''' ),
        __.tyro.conf.arg( aliases = ( '-e', '--edit-message' ) ),
    ] = False
    prepend_prompt: __.typx.Annotated[
        bool,
        __.typx.Doc( ''' Prepend mimeogram format instructions. ''' ),
    ] = False
    recurse: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.typx.Doc( ''' Recurse into directories. ''' ),
        __.tyro.conf.arg(
            aliases = ( '-r', '--recurse-directories', '--recursive' ) ),
    ] = None
    strict: __.typx.Annotated[
        __.typx.Optional[ bool ],
        __.typx.Doc(
            ''' Fail on invalid contents? True, fail. False, skip. ''' ),
        __.tyro.conf.arg( aliases = ( '--fail-on-invalid', ) ),
    ] = None
    tokenizer: __.typx.Annotated[
        __.typx.Optional[ _tokenizers.Tokenizers ],
        __.typx.Doc( ''' Which tokenizer to use for counting? ''' ),
    ] = None
    tokenizer_variant: __.typx.Annotated[
        __.typx.Optional[ str ],
        __.typx.Doc(
            ''' Which tokenizer variant to use for counting?

                'tiktoken': 'cl100k_base', 'o200k_base', etc....

                Not all tokenizers have variants.
                If not specified, then the default variant is used.
            ''' ),
    ] = None

    async def __call__( self, auxdata: __.Globals ) -> None:
        ''' Executes command to create mimeogram. '''
        await create( auxdata, self )

    def provide_configuration_edits( self ) -> __.DictionaryEdits:
        ''' Provides edits against configuration from options. '''
        edits: list[ __.DictionaryEdit ] = [ ]
        if None is not self.clip:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'create', 'to-clipboard' ), value = self.clip ) )
        if None is not self.count_tokens:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'create', 'count-tokens' ),
                value = self.count_tokens ) )
        if None is not self.recurse:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'acquire-parts', 'recurse-directories' ),
                value = self.recurse ) )
        if None is not self.strict:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'acquire-parts', 'fail-on-invalid' ),
                value = self.strict ) )
        if None is not self.tokenizer:
            edits.append( __.SimpleDictionaryEdit( # pyright: ignore
                address = ( 'tokenizers', 'default' ),
                value = self.tokenizer ) )
        return tuple( edits )


async def _acquire_prompt( auxdata: __.Globals ) -> str:
    from .prompt import acquire_prompt
    return await acquire_prompt( auxdata )


async def _copy_to_clipboard( mimeogram: str ) -> None:
    from pyperclip import copy
    copy( mimeogram )
    _scribe.info( "Copied mimeogram to clipboard." )


async def _edit_message( ) -> str:
    from .edit import edit_content
    return edit_content( )


async def create( # pylint: disable=too-complex,too-many-locals
    auxdata: __.Globals,
    command: Command, *,
    editor: __.cabc.Callable[
        [ ], __.cabc.Coroutine[ None, None, str ] ] = _edit_message,
    clipcopier: __.cabc.Callable[
        [ str ], __.cabc.Coroutine[ None, None, None ] ] = _copy_to_clipboard,
    prompter: __.cabc.Callable[
        [ __.Globals ],
        __.cabc.Coroutine[ None, None, str ] ] = _acquire_prompt,
) -> __.typx.Never:
    ''' Creates mimeogram. '''
    from .acquirers import acquire
    from .formatters import format_mimeogram
    with __.report_exceptions(
        _scribe, "Could not acquire mimeogram parts."
    ): parts = await acquire( auxdata, command.sources )
    if command.edit:
        with __.report_exceptions(
            _scribe, "Could not acquire user message."
        ): message = await editor( )
    else: message = None
    mimeogram = format_mimeogram( parts, message = message )
    # TODO? Pass prompt to 'format_mimeogram'.
    if command.prepend_prompt:
        prompt = await prompter( auxdata )
        mimeogram = f"{prompt}\n\n{mimeogram}"
    options = auxdata.configuration.get( 'create', { } )
    if options.get( 'count-tokens', False ):
        with __.report_exceptions(
            _scribe, "Could not count mimeogram tokens."
        ):
            tokenizer = await _tokenizer_from_command( auxdata, command )
            tokens_count = await tokenizer.count( mimeogram )
            _scribe.info( f"Total mimeogram size is {tokens_count} tokens." )
    if options.get( 'to-clipboard', False ):
        with __.report_exceptions(
            _scribe, "Could not copy mimeogram to clipboard."
        ): await clipcopier( mimeogram )
    else: print( mimeogram ) # TODO? Use output stream from configuration.
    raise SystemExit( 0 )


async def _tokenizer_from_command(
    auxdata: __.Globals, command: Command
) -> _tokenizers.Tokenizer:
    options = auxdata.configuration.get( 'tokenizers', { } )
    name = (
        command.tokenizer.value if command.tokenizer
        else options.get( 'default', 'tiktoken' ) )
    variant = command.tokenizer_variant
    args = dict( variant = variant ) if variant else { }
    return await _tokenizers.Tokenizers.produce( name, **args )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Family of exceptions for package internals.

    * ``Omniexception``: Base for all internal exceptions
    * ``Omnierror``: Base for all internals errors
'''


import contextlib as _contextlib
import logging as _logging

import exceptiongroup as _exceptiongroup

from . import imports as __


class Omniexception( BaseException, metaclass = __.ImmutableClass ):
    ''' Base for all exceptions raised internally. '''
    # TODO: Class attribute concealment.
    # TODO: Instance attribute concealment and immutability.

    _attribute_visibility_includes_: __.cabc.Collection[ str ] = (
        frozenset( ( '__cause__', '__context__', ) ) )


class Omnierror( Omniexception, Exception ):
    ''' Base for error exceptions raised internally. '''


class AddressLocateFailure( Omnierror, LookupError ):
    ''' Failure to locate address. '''

    def __init__(
        self, subject: str, address: __.cabc.Sequence[ str ], part: str
    ):
        super( ).__init__(
            f"Could not locate part '{part}' of address '{address}' "
            f"in {subject}." )


class AsyncAssertionFailure( Omnierror, AssertionError, TypeError ):
    ''' Assertion of awaitability of entity failed. '''

    def __init__( self, entity: __.typx.Any ):
        super( ).__init__( f"Entity must be awaitable: {entity!r}" )


class EntryAssertionFailure( Omnierror, AssertionError, KeyError ):
    ''' Assertion of entry in dictionary failed. '''

    def __init__( self, subject: str, name: str ):
        super( ).__init__( f"Could not find entry '{name}' in {subject}." )


class OperationInvalidity( Omnierror, RuntimeError ):
    ''' Invalid operation. '''

    def __init__( self, subject: str, name: str ):
        super( ).__init__(
            f"Could not perform operation '{name}' on {subject}." )


@_contextlib.contextmanager
def report_exceptions(
    scribe: _logging.Logger,
    message: str,
    eclass: type[ BaseException ] = SystemExit,
    eposargs: __.cabc.Sequence[ __.typx.Any ] = ( 1, ),
) -> __.cabc.Generator[ None, None, None ]:
    ''' Intercepts and reports exceptions.

        By default, raises ``SystemExit( 1 )``.
    '''
    level = scribe.getEffectiveLevel( )
    try: yield
    except _exceptiongroup.ExceptionGroup as excg: # pyright: ignore
        scribe.error( message ) # noqa: TRY400
        for exc in excg.exceptions: # pyright: ignore
            if level <= _logging.DEBUG:
                nomargs = dict( exc_info = exc ) # pyright: ignore
            else: nomargs = { }
            scribe.error( # noqa: TRY400
                f"\tCause: {exc}", **nomargs ) # pyright: ignore
        if eclass: raise eclass( *eposargs ) from None
    except Exception as exc: # pylint: disable=broad-exception-caught
        if level <= _logging.DEBUG: scribe.exception( f"{message}" )
        else: scribe.error( f"{message} Cause: {exc}" ) # noqa: TRY400
        if eclass: raise eclass( *eposargs ) from None



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Language model tokenizers. '''


from __future__ import annotations

import tiktoken as _tiktoken

from . import __


_scribe = __.produce_scribe( __name__ )



class Tokenizers( __.enum.Enum ):
    ''' Language model tokenizers. '''

    AnthropicApi =  'anthropic-api'
    Tiktoken =      'tiktoken'

    @classmethod
    async def produce(
        selfclass, name: str, variant: __.Absential[ str ] = __.absent
    ) -> Tokenizer:
        ''' Produces tokenizer from name and optional variant. '''
        tokenizer = selfclass( name )
        match tokenizer:
            case Tokenizers.AnthropicApi:
                raise NotImplementedError( # noqa: TRY003
                    "Not implemented yet. Sorry." )
            case Tokenizers.Tiktoken:
                return await Tiktoken.from_variant( name = variant )


# pylint: disable=invalid-metaclass
class Tokenizer(
    __.typx.Protocol,
    metaclass = __.ImmutableStandardProtocolDataclass,
    decorators = ( __.standard_dataclass, __.typx.runtime_checkable, ),
):
    ''' Language model tokenizer. '''

    @classmethod
    @__.abc.abstractmethod
    async def from_variant(
        selfclass, name: __.Absential[ str ] = __.absent
    ) -> __.typx.Self:
        ''' Produces instance from name of variant. '''

    @__.abc.abstractmethod
    async def count( self, text: str ) -> int:
        ''' Counts number of tokens in text. '''
        raise NotImplementedError
# pylint: enable=invalid-metaclass


# TODO: Implement 'AnthropicApi' tokenizer.


class Tiktoken(
    Tokenizer, decorators = ( __.standard_dataclass, )
):
    ''' Tokenization via 'tiktoken' package. '''

    codec: _tiktoken.Encoding

    @classmethod
    async def from_variant(
        selfclass, name: __.Absential[ str ] = __.absent
    ) -> __.typx.Self:
        if __.is_absent( name ): name = 'cl100k_base'
        from tiktoken import get_encoding
        try: codec = get_encoding( name )
        except ValueError as exc:
            from .exceptions import TokenizerVariantInvalidity
            raise TokenizerVariantInvalidity( 'tiktoken', name ) from exc
        return selfclass( codec = codec )

    async def count( self, text: str ) -> int:
        return len( self.codec.encode( text ) )
